// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

use datafusion::arrow::datatypes::DataType;
use pyo3::prelude::*;

use crate::common::data_type::PyDataType;

#[pyclass(name = "Placeholder", module = "datafusion.expr", subclass)]
#[derive(Clone)]
pub struct PyPlaceholder {
    id: String,
    data_type: Option<DataType>,
}

impl PyPlaceholder {
    pub fn new(id: String, data_type: DataType) -> Self {
        Self {
            id,
            data_type: Some(data_type),
        }
    }
}

#[pymethods]
impl PyPlaceholder {
    fn id(&self) -> String {
        self.id.clone()
    }

    fn data_type(&self) -> Option<PyDataType> {
        self.data_type.as_ref().map(|e| e.clone().into())
    }
}
